CodeQL / Analyze (actions) (push) Canceled after 0s
CodeQL / Analyze (python) (push) Canceled after 0s
pre-commit / pre-commit (push) Canceled after 0s
Multi-Arch CI / setup (push) Canceled after 0s
Publish build_manylinux_x86_64 images / publish_build_manylinux_x86_64 (push) Canceled after 0s
Setup (Multi-Arch) / setup (push) Canceled after 0s
Multi-Arch CI / Linux::${{ fromJSON(needs.setup.outputs.linux_build_config || '{}').build_variant_label || 'skip' }} (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish no_rocm_image_ubi10 image / publish_no_rocm_image_ubi10 (push) Canceled after 0s
Multi-Arch CI / Windows::${{ fromJSON(needs.setup.outputs.windows_build_config || '{}').build_variant_label || 'skip' }} (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Multi-Arch CI / Manifest Diff (push) Canceled after 0s
TheRock Manifest Diff Report / Generate Manifest Diff Report (push) Canceled after 0s
Multi-Arch CI / CI Summary (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04 (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04_ocl_rt images / publish_no_rocm_image_ubuntu24_04_ocl_rt (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_rocgdb (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04_rocgdb images / publish_no_rocm_image_ubuntu24_04_rocgdb (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_ocl_rt (push) Canceled after 0s
Publish TheRock Docker image / build-and-push-image (push) Canceled after 0s
Unit Tests / Unit Tests :: ubuntu-24.04 (push) Canceled after 0s
Unit Tests / Unit Tests :: windows-2022 (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_media (push) Canceled after 0s
Unit Tests / Unit Tests Summary (push) Canceled after 0s
Publish no_rocm_image_ubuntu24_04 images / publish_no_rocm_image_ubuntu24_04_openmpi (push) Canceled after 0s
93 lines
3.7 KiB
Plaintext
93 lines
3.7 KiB
Plaintext
// Copyright (c) 2020-2025 Advanced Micro Devices, Inc. All rights reserved.
|
|
//
|
|
// Permission is hereby granted, free of charge, to any person obtaining a copy
|
|
// of this software and associated documentation files (the "Software"), to deal
|
|
// in the Software without restriction, including without limitation the rights
|
|
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
|
|
// copies of the Software, and to permit persons to whom the Software is
|
|
// furnished to do so, subject to the following conditions:
|
|
//
|
|
// The above copyright notice and this permission notice shall be included in all
|
|
// copies or substantial portions of the Software.
|
|
//
|
|
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
|
|
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
|
|
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
|
|
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
|
|
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
|
|
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
|
|
// SOFTWARE.
|
|
|
|
#include <thrust/detail/config.h>
|
|
|
|
#include <thrust/device_vector.h>
|
|
#include <thrust/reduce.h>
|
|
#include <thrust/system/cuda/execution_policy.h>
|
|
|
|
#include <cassert>
|
|
#include <future>
|
|
|
|
// This example demonstrates two ways to achieve algorithm invocations that are asynchronous with
|
|
// the calling thread.
|
|
//
|
|
// The first method wraps a call to thrust::reduce inside a __global__ function. Since __global__ function
|
|
// launches are asynchronous with the launching thread, this achieves asynchrony. The result of the reduction
|
|
// is stored to a pointer to CUDA global memory. The calling thread waits for the result of the reduction to
|
|
// be ready by synchronizing with the CUDA stream on which the __global__ function is launched.
|
|
//
|
|
// The second method uses the C++11 library function, std::async, to create concurrency. The lambda function
|
|
// given to std::async returns the result of thrust::reduce to a std::future. The calling thread can use the
|
|
// std::future to wait for the result of the reduction. This method requires a compiler which supports
|
|
// C++11-capable language and library constructs.
|
|
|
|
template <typename Iterator, typename T, typename BinaryOperation, typename Pointer>
|
|
__global__ void reduce_kernel(Iterator first, Iterator last, T init, BinaryOperation binary_op, Pointer result)
|
|
{
|
|
*result = thrust::reduce(thrust::cuda::par, first, last, init, binary_op);
|
|
}
|
|
|
|
int main()
|
|
{
|
|
size_t n = 1 << 20;
|
|
thrust::device_vector<unsigned int> data(n, 1);
|
|
thrust::device_vector<unsigned int> result(1, 0);
|
|
|
|
// method 1: call thrust::reduce from an asynchronous CUDA kernel launch
|
|
|
|
// create a CUDA stream
|
|
cudaStream_t s;
|
|
cudaStreamCreate(&s);
|
|
|
|
// launch a CUDA kernel with only 1 thread on our stream
|
|
reduce_kernel<<<1, 1, 0, s>>>(data.begin(), data.end(), 0, thrust::plus<int>(), result.data());
|
|
|
|
// wait for the stream to finish
|
|
cudaStreamSynchronize(s);
|
|
|
|
// our result should be ready
|
|
assert(result[0] == n);
|
|
|
|
cudaStreamDestroy(s);
|
|
|
|
// reset the result
|
|
result[0] = 0;
|
|
|
|
// method 2: use std::async to create asynchrony
|
|
// copy all the algorithm parameters
|
|
auto begin = data.begin();
|
|
auto end = data.end();
|
|
unsigned int init = 0;
|
|
auto binary_op = thrust::plus<unsigned int>();
|
|
|
|
// std::async captures the algorithm parameters by value
|
|
// use std::launch::async to ensure the creation of a new thread
|
|
std::future<unsigned int> future_result = std::async(std::launch::async, [=] {
|
|
return thrust::reduce(begin, end, init, binary_op);
|
|
});
|
|
|
|
// wait on the result and check that it is correct
|
|
assert(future_result.get() == n);
|
|
|
|
return 0;
|
|
}
|